#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solubility package"""

# Expose the public API
from atmodeller.solubility.core import (  # noqa: E402, F401
    NoSolubility,
    Solubility,
    SolubilityPowerLaw,
    SolubilityPowerLawLog10,
)
from atmodeller.solubility.library import get_solubility_models  # noqa: E402, F401
